from vqr.cvqf import DiscreteCVQF
from vqr.solvers.cvx import CVXVQRSolver
from vqr.solvers.pot import POTVQESolver
from vqr.solvers.base import Solver, VQESolver, VQRSolver, DiscreteSolver
from vqr.solvers.regularized_lse import (
    RegularizedDualVQRSolver,
    MLPRegularizedDualVQRSolver,
)
